import datetime
from functools import partial
from logging import Logger
from typing import Iterator, Union

from lime_trader.api.authenticated_api_client import AuthenticatedApiClient
from lime_trader.models.accounts import AccountTransaction
from lime_trader.models.market import (Quote, QuoteHistory, Period, Security, SecuritiesPage, Trade, TradesPage,
                                       CurrentSchedule, OptionSeries, OptionChain)
from lime_trader.models.page import Page, PageRequest
from lime_trader.constants.urls import (MARKET_DATA_GET_CURRENT_QUOTE, MARKET_DATA_GET_QUOTES,
                                        MARKET_DATA_GET_TRADING_SCHEDULE,
                                        MARKET_DATA_LOOKUP_SECURITIES, MARKET_DATA_GET_TIME_AND_SALES,
                                        MARKET_DATA_GET_QUOTES_HISTORY, MARKET_DATA_GET_OPTION_CHAIN,
                                        MARKET_DATA_GET_OPTION_SERIES)
from lime_trader.utils.pagination import iterate_pages


class MarketDataClient:
    """
    Contains methods related to market data
    """

    def __init__(self, api_client: AuthenticatedApiClient, logger: Logger):
        """
        Args:
            api_client: API client that will be used to execute all requests
            logger: Logger used to submit client log messages
        """
        self._api_client = api_client
        self._logger = logger

    def get_current_quote(self, symbol: str) -> Quote:
        """
        Retrieves current quote for the specified symbol

        Args:
            symbol: Security symbol

        Returns:
            Quote for the symbol
        """
        self._logger.info(f"Getting current quote for symbol {symbol}")
        return self._api_client.get(MARKET_DATA_GET_CURRENT_QUOTE, path_params={},
                                    params={"symbol": symbol}, response_schema=Quote)

    def get_current_quotes(self, symbols: list[str]) -> list[Quote]:
        """
        Retrieves current quotes for all specified symbols

        Args:
            symbols: List of security symbols

        Returns:
            List of quotes for specified symbols
        """
        self._logger.info(f"Getting current quotes for symbols {symbols}")
        return self._api_client.post(MARKET_DATA_GET_QUOTES, path_params={}, json=symbols, response_schema=list[Quote])

    def get_quotes_history(self, symbol: str, period: Period, from_date: datetime.datetime,
                           to_date: datetime.datetime) -> list[QuoteHistory]:
        """
        Returns candle structures aggregated by specified period

        Args:
            symbol: The security symbol, stocks in Nasdaq CMS convention. Options in OCC
            period: Aggregation period.
            from_date: Period start
            to_date: Period end

        Returns:
            Quote history between specified period
        """
        self._logger.info(f"Getting quotes history for symbol={symbol}, period={period}, from_date={from_date},"
                          f"to_date={to_date}")
        return self._api_client.get(MARKET_DATA_GET_QUOTES_HISTORY, path_params={},

                                    params={"symbol": symbol,
                                            "period": period.value,
                                            "from": from_date,
                                            "to": to_date
                                            }, response_schema=list[QuoteHistory])

    def get_trading_schedule(self) -> CurrentSchedule:
        """
        Returns trading session info depending on current date and time:

        Returns:
            Current trading schedule
        """
        self._logger.info(f"Getting current trading schedule")
        return self._api_client.get(MARKET_DATA_GET_TRADING_SCHEDULE, path_params={}, params={},
                                    response_schema=CurrentSchedule)

    def lookup_securities(self, query: str, page: PageRequest) -> Page[Security]:
        """
        Searches the securities reference by the specified criteria. The criteria can be a symbol,
        part of a symbol or part of a company name.

        Args:
            query: Search criteria
            page: Pagination parameters

        Returns:
            Page of security symbols which match criteria
        """
        self._logger.info(f"Looking up securities with query={query}, page={page}")
        response = self._api_client.get(MARKET_DATA_LOOKUP_SECURITIES, path_params={},
                                        params={"query": query,
                                                "limit": page.size,
                                                "skip": page.get_offset()},
                                        response_schema=SecuritiesPage)
        return Page(data=response.securities, number=page.page, size=page.size, total_elements=response.count)

    def time_and_sales(self, symbol: str, from_date: datetime.datetime,
                       to_date: datetime.datetime, page: PageRequest) -> Page[Trade]:
        """
        Retrieves the time and sales history, ordered by descending timestamp. Query parameters are:

        Args:
            symbol: The security symbol, stocks in Nasdaq CMS convention. Options are not supported
            from_date: Period start
            to_date: Period end
            page: Pagination parameters

        Returns:
            Page of historical trades
        """
        self._logger.info(f"Getting time and sales history for symbol={symbol}, from_date={from_date},"
                          f"to_date={to_date}, page={page}")
        response = self._api_client.get(MARKET_DATA_GET_TIME_AND_SALES, path_params={},
                                        params={"symbol": symbol,
                                                "from": from_date,
                                                "to": to_date,
                                                "limit": page.size,
                                                "skip": page.get_offset()},
                                        response_schema=TradesPage)
        return Page(data=response.trades, number=page.page, size=page.size, total_elements=response.count)

    def iterate_time_and_sales(self, symbol: str, from_date: datetime.datetime,
                               to_date: datetime.datetime, start_page: PageRequest,
                               ) -> Iterator[Page[AccountTransaction]]:
        """
        Returns iterator for the time and sales history, ordered by descending timestamp

        Args:
            symbol: The security symbol, stocks in Nasdaq CMS convention. Options are not supported
            from_date: Period start
            to_date: Period end
            start_page: Start page. Iterator starts from this page

        Returns:
            Iterator of pages of historical trades
        """
        for page in iterate_pages(start_page=start_page,
                                  func=partial(self.time_and_sales, symbol=symbol,
                                               from_date=from_date, to_date=to_date)):
            yield page

    def get_option_series(self, symbol: str) -> list[OptionSeries]:
        """
        Returns an array of option series by the specified underlying security's symbol.
        Contains at least one element which is the default series.
        Option Series is a specific set of calls or puts on the same underlying security,
        with the same strike price and expiration date.
        For the most part there is just one series name that matches the symbol but in some cases related to
        corporate actions on the underlying security additional series are issued.

        Args:
            symbol: The security symbol
        Returns:
            Array of option series by the specified underlying security's symbol
        """
        return self._api_client.get(MARKET_DATA_GET_OPTION_SERIES, path_params={"symbol": symbol}, params={},
                                    response_schema=list[OptionSeries])

    def get_option_chain(self, symbol: str, expiration: datetime.date,
                         series: Union[str, None] = None) -> OptionChain:
        """
        Returns option contracts for specified symbol, expiration date and series.

        Args:
            symbol: The security symbol.
            expiration: Contract expiration date
            series: By default the series is the same as the security symbol

        Returns:
            Option contracts for specified symbol, expiration date and series
        """
        return self._api_client.get(MARKET_DATA_GET_OPTION_CHAIN, path_params={"symbol": symbol},
                                    params={"expiration": expiration, "series": series},
                                    response_schema=OptionChain)



import logging
from json import JSONDecodeError
from logging import Logger
from typing import TypeVar, Type, Any, Callable
import httpx
import websocket
from httpx import Response, HTTPStatusError
from httpx._client import BaseClient
from urllib3.util import Url

from lime_trader.converters.abstract_converter import AbstractConverter
from lime_trader.models.errors import Error
from lime_trader.exceptions.api_error import ApiError
from lime_trader.utils.logging_utils import log_json_data

T = TypeVar("T")


class ApiClient:
    """
    Used for communication with the API
    """

    def __init__(self, base_url: Url, converter: AbstractConverter, use_https: bool, use_wss: bool,
                 timeout: int,
                 http_client: BaseClient,
                 logger: Logger):
        """
        Args:
            base_url: Base url that will be appended to all urls
            converter: Converter to use for transforming response data to Python objects
            use_https: Indicator if https protocol should be used, http is used if False
            use_wss: Indicator if wss protocol should be used, ws is used if False
            timeout: Request timeout
            logger: Logger to use for logging messages
        """
        self._base_url = base_url
        self._converter = converter
        self._use_https = use_https
        self._use_wss = use_wss
        self._timeout = timeout
        self._logger = logger
        self._http_client = http_client

    def _map_response(self, data: dict[str, Any] | list[str, Any], t: Type[T]) -> T:
        """
        Does conversion between API response (json converted to dictionary) and Python type

        Args:
            data: Response data that needs to be converted
            t: Type to convert to

        Returns:
            Data converted to Python type passed as t parameter
        """
        return self._converter.load_from_dict(data, t)

    def _check_for_errors(self, response: Response) -> dict[str, Any]:
        """
        Checks for errors in a response. If there is, raises exception with returned error

        Args:
            response: Response to check for errors

        Returns:
            True if there is no error

        Raises:
            ApiError if response does not have success http status code
        """
        try:
            data = response.json()
        except JSONDecodeError as e:
            raise ApiError(
                Error(status_code=response.status_code,
                      code="json_decode_error",
                      message=e.msg)
            )
        try:
            response.raise_for_status()
        except HTTPStatusError as e:
            raise ApiError(
                Error(status_code=response.status_code,
                      code=data.get("code", ""),
                      message=data.get("message", ""))
            )

        return data

    def _handle_response(self, response: Response, response_schema: Type[T]) -> T:
        """
        Handles response by checking for errors and transforming data to appropriate Python type

        Args:
            response: Response returned from API
            response_schema: Python type to which response data will be transformed

        Returns:
            Pythonic type
        """
        response_data = self._check_for_errors(response)
        self._logger.debug(f"Response: {log_json_data(response_data)}")
        return self._map_response(response_data, response_schema)

    def get(self, url: Url, path_params: dict[str, Any], params: Any, headers: dict[str, str],
            response_schema: Type[T]) -> T:
        """
        Executes GET request with specified parameters

        Args:
            url: Url path.
            path_params: URL path parameters
            params: GET parameters
            headers: Request headers
            response_schema: Python type to which response data will be transformed

        Returns:
            Response converted to Python type passed as response_schema parameter
        """
        request_url = self._get_url(url, path_params=path_params)
        dict_params = self._converter.dump_to_dict(params, remove_none=True)

        self._logger.debug(f"Executing GET request {request_url}, params={log_json_data(dict_params)}")
        try:
            response = self._http_client.get(url=request_url, params=dict_params, headers=headers,
                                             timeout=self._timeout)

        except httpx.TransportError as e:
            raise ApiError(
                Error(status_code=None,
                      code="transport_error",
                      message=f"Transport error: {e.__class__.__name__}.{''.join(e.args)}")
            )
        return self._handle_response(response=response, response_schema=response_schema)

    def _get_url(self, path: Url, path_params: dict[str, Any]) -> str:
        """
        Gets full url by prefixing base url to path

        Args:
            path: Path to append to base url.
            path_params: URL path parameters

        Returns:
            Full url
        """
        if path.url.startswith("http://"):
            scheme = "https" if self._use_https else "http"
        elif path.url.startswith("ws://"):
            scheme = "wss" if self._use_wss else "ws"
        else:
            raise Exception("Invalid urls specified!")
        url = Url(scheme=scheme, path=(self._base_url.path or '') + path.path, host=self._base_url.hostname).url
        return url.format(**self._converter.dump_to_dict(path_params))

    def post_form(self, url: Url, path_params: dict[str, Any], data: Any, headers: dict[str, str],
                  response_schema: Type[T]) -> T:
        """
        Executes POST request with form data

        Args:
            url: Url path.
            path_params: URL path parameters
            data: Form data
            headers: Request headers
            response_schema: Python type to which response data will be transformed

        Returns:
            Response converted to Python type passed as response_schema parameter
        """
        request_url = self._get_url(url, path_params=path_params)
        dict_data = self._converter.dump_to_dict(data, remove_none=True)
        self._logger.debug(f"Executing POST request {request_url}, data={log_json_data(dict_data)}")
        try:
            response = self._http_client.post(url=request_url, data=dict_data, headers=headers, timeout=self._timeout)
        except httpx.TransportError as e:
            raise ApiError(
                Error(status_code=None,
                      code="transport_error",
                      message=f"Transport error: {e.__class__.__name__}.{''.join(e.args)}")
            )
        result = self._handle_response(response=response, response_schema=response_schema)
        return result

    def _prepare_json_body(self, json: Any) -> dict[str, Any] | list[dict[str, Any]]:
        if type(json) == list:
            return [self._converter.dump_to_dict(item, remove_none=True) for item in json]
        return self._converter.dump_to_dict(json, remove_none=True)

    def post(self, url: Url, path_params: dict[str, Any], json: Any, headers: dict[str, str],
             response_schema: Type[T]) -> T:
        """
        Executes POST request with JSON data

        Args:
            url: Url path.
            path_params: URL path parameters
            json: JSON data
            headers: Request headers
            response_schema: Python type to which response data will be transformed

        Returns:
            Response converted to Python type passed as response_schema parameter
        """
        request_url = self._get_url(url, path_params=path_params)
        dict_data = self._prepare_json_body(json)
        self._logger.debug(f"Executing POST request {request_url}, json={log_json_data(dict_data)}")
        try:
            response = self._http_client.post(url=request_url, json=dict_data, headers=headers, timeout=self._timeout)
        except httpx.TransportError as e:
            raise ApiError(
                Error(status_code=None,
                      code="transport_error",
                      message=f"Transport error: {e.__class__.__name__}.{''.join(e.args)}")
            )
        return self._handle_response(response=response, response_schema=response_schema)

    def websocket_connection(self, url: Url, path_params: dict[str, Any], on_message: Callable,
                             on_error: Callable,
                             headers: dict[str, str]) -> websocket.WebSocketApp:
        request_url = self._get_url(url, path_params=path_params)
        self._logger.debug(f"Opening websocket connection {request_url}")
        if self._logger.level <= logging.DEBUG:
            websocket.enableTrace(True)
        ws = websocket.WebSocketApp(url=request_url,
                                    on_message=on_message, on_error=on_error,
                                    header=headers)

        return ws



import time

from httpx import Request
from httpx._client import logger


def log_request(request: Request):
    request.start_time = time.time()


def log_response(response):
    request = response.request
    time_elapsed = time.time() - response.request.start_time
    logger.debug(
        f"Response event hook: {request.method} {request.url} - Status {response.status_code}, elapsed={time_elapsed * 1000}ms")

async def async_log_request(request: Request):
    request.start_time = time.time()


async def async_log_response(response):
    request = response.request
    time_elapsed = time.time() - response.request.start_time
    logger.debug(
        f"Response event hook: {request.method} {request.url} - Status {response.status_code}, elapsed={time_elapsed * 1000}ms")
